#!/usr/bin/env python
# $Id: fetchhandler.py 3928 2008-12-13 02:12:34Z ltsai $
#
# Copyright 2008 Platform Computing Inc.
#
import base64
import shutil
import socket
import urllib2
import urlparse
import subprocess
from path import path
from primitive.fetchtool.helper import copyDir
from primitive.fetchtool.helper import isDestWritable
from primitive.fetchtool.helper import decipherWgetErrors
from primitive.fetchtool.helper import checkUserPassFormat
from primitive.core.errors import HTTPAuthException
from primitive.core.errors import UnknownHostException
from primitive.core.errors import FetchException

# urlparse by default does not recognise
# these protocols, so we add them.
for scheme in ('rhn', 'rhns', 'you', 'scp'):
    urlparse.uses_netloc.insert(0, scheme)
    urlparse.uses_query.insert(0, scheme)


class ProtocolFetchHandler(object):
    def __init__(self, **kwargs):
        self.required_args = []

    def fetch(self, **kwargs):
        raise NotImplementedError


class FileFetchHandler(ProtocolFetchHandler):
    ''' FileFetchHandler class is a subclass of FetchHandler which handles file
        transfers local to the system, e.g., copying from a dvd mounted at
        /media/dvd onto /data/foo. It can copy a single file or a directory
        to the specified destination directory.
    '''
    def __init__(self, **kwargs):
        ProtocolFetchHandler.__init__(self, **kwargs)

    def fetch(self, **kwargs):
        ''' Check if we are fetching a single file or
            a directory and act accordingly.
        '''
        if kwargs['fetchdir']:
            dest = self.copyDir(kwargs['uri'],
                                kwargs['destdir'],
                                kwargs['overwrite'])
        else:
            dest = self.copyFile(kwargs['uri'],
                                 kwargs['destdir'])
        return (True, dest)

    def copyFile(self, src, dest_dir):
        ''' Copy only one file from the src to destdir. '''
        dest = (path(dest_dir) / path(src).basename()).realpath()
        src_file = path(urlparse.urlparse(src)[2]).realpath()
        if src_file != dest:
            shutil.copy2(src_file, dest)
        return dest 

    def copyDir(self, src_uri, dest_dir, overwrite):
        ''' Fetch the srcdir to destdir. '''
        src_dir = urlparse.urlparse(src_uri)[2]
        # This is a call to the global copydir defined in the helper.
        copyDir(src_dir, dest_dir,
                recursive=True,
                overwrite=overwrite)
        return dest_dir


class HTTPFetchHandler(ProtocolFetchHandler):
    ''' HTTPFetchHandler class is a subclass of FetchHandler which fetches file(s)
        from a HTTP location to a directory local to the system, e.g.,
        downloading "http://www.foo.com/foo.txt" as "foo.txt" in the local
        output directory, or mirroring "http://www.bar.com/foodir/" directory
        structure on the local filesystem.
    '''
    def __init__(self, **kwargs):
        ProtocolFetchHandler.__init__(self, **kwargs)

    def fetch(self, **kwargs):
        ''' Check if we are fetching a single file or
            mirroring a directory and act accordingly.
        '''
        if not checkUserPassFormat(kwargs['uri']):
            raise InvalidUserPassPairException, 'Invalid username/password given.'

        self.validateArgs(**kwargs)

        if kwargs['fetchdir']:
            dest = self.mirrorDir(kwargs['uri'],
                           kwargs['destdir'])
        else:
            dest = self.downloadFile(kwargs['uri'],
                              kwargs['destdir'])
        return (True, dest)

    def validateArgs(self, **kwargs):
        ''' Override the validate input function with a
            few more checks.
        '''
        # check that the hostname is reachable.
        host = urlparse.urlparse(kwargs['uri'])[1]
        userhost = host.split('@')
        try:
            socket.gethostbyname(userhost[-1])
        except socket.gaierror:
            raise UnknownHostException

        # ensure the username and password pair are correct.
        req = urllib2.Request('http://' + userhost[-1] + \
                              urlparse.urlparse(kwargs['uri'])[2])
        if len(userhost) > 1:
            # Didn't use base64.encodestring(). Reason: had '\n' in the middle of the string.
            base64string = base64.b64encode(userhost[0])
            authheader = 'Basic %s' % base64string
            req.add_header('Authorization', authheader)
            try:
                urllib2.urlopen(req)
            except urllib2.HTTPError, e:
                if hasattr(e, 'code') and e.code == 401:
                    raise HTTPAuthException

    def downloadFile(self, src_uri, dest_dir):
        ''' Download only one file from the src to destdir. '''
        if not src_uri.endswith('/'):
            dest_filename = path(urlparse.urlsplit(src_uri)[2]).basename()
        else:
            dest_filename = path(urlparse.urlsplit(src_uri[-1])[2]).basename()
        if not dest_filename:
            raise FetchException,'Unable to obtain a destination file to download'
        dest_path = path(dest_dir) / dest_filename
        if dest_path.exists():
            dest_path.remove()
        p = subprocess.Popen(['wget', '--no-verbose', src_uri,
                              '--directory-prefix', dest_dir],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        retcode = p.wait()
        stderr = p.stderr.readlines()
        if retcode:
            #Raises FetchException on errors.
            decipherWgetErrors(src_uri, stderr)
        return path(path(dest_dir) / dest_filename)

    def mirrorDir(self, src_uri, dest_dir):
        ''' Mirror a remote HTTP directory.
            calls 'wget --no-parent -cut-dirs=XX -nH --mirror'.
        '''
        p = path(urlparse.urlsplit(src_uri)[2]).splitall()
        # Deals with non-ending slash.
        # Non-ending clash url ends up with an empty string in the
        # last index of the list when a splitall is done
        if not p[-1]:
            cutaway = len(p[1:]) - 1
        else: # non-ending slash
            cutaway = len(p[1:])
            src_uri = src_uri + '/' # Append a trailing slash
        if cutaway < 0:
            cutaway = 0

        wget = subprocess.Popen(['wget', '--no-verbose', '--no-parent',
                                '--cut-dirs=%s' % cutaway, '-nH', 
                                '--mirror', src_uri,
                                '--directory-prefix', dest_dir],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        retcode = wget.wait()
        decipherWgetErrors(src_uri, wget.stderr.readlines())
        return dest_dir


class HTTPSFetchHandler(HTTPFetchHandler):
    ''' HTTPSFetchHandler class subclasses from HTTPFetchHandler, and has
        identical function to its superclass, except that it handles the
        https protocol
    '''
    def __init__(self, **kwargs):
        HTTPFetchHandler.__init__(self, **kwargs)

    def validateArgs(self, **kwargs):
        ''' Override the validate input function with a
            few more checks.
        '''
        # check that the hostname is reachable.
        host = urlparse.urlparse(kwargs['uri'])[1]
        userhost = host.split('@')
        try:
            socket.gethostbyname(userhost[-1])
        except socket.gaierror:
            raise UnknownHostException

        # ensure the username and password pair are correct.
        req = urllib2.Request('https://' + userhost[-1] + \
                              urlparse.urlparse(kwargs['uri'])[2])
        if len(userhost) > 1:
            # Didn't use base64.encodestring(). Reason: had '\n' in the middle of the string.
            base64string = base64.b64encode(userhost[0])
            authheader = 'Basic %s' % base64string
            req.add_header('Authorization', authheader)
            try:
                urllib2.urlopen(req)
            except urllib2.HTTPError, e:
                if hasattr(e, 'code') and e.code == 401:
                    raise HTTPAuthException

    def downloadFile(self, src_uri, dest_dir):
        ''' Download only one file from the src to destdir. '''
        # Obtain the destination filename after stripping away trailing Slash
        if not src_uri.endswith('/'):
            dest_filename = path(urlparse.urlsplit(src_uri)[2]).basename()
        else:
            dest_filename = path(urlparse.urlsplit(src_uri[-1])[2]).basename()
        if not dest_filename:
            raise FetchException,'Unable to obtain a destination file to download'
       
        dest_path = path(dest_dir) / dest_filename
        if dest_path.exists():
            dest_path.remove()
        p = subprocess.Popen(['wget', '--no-verbose',
                              '--no-check-certificate', src_uri,
                              '--directory-prefix', dest_dir],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        retcode = p.wait()
        stderr = p.stderr.readlines()
        if retcode:
            decipherWgetErrors(src_uri, stderr)
        return path(path(dest_dir) / dest_filename)
        
        

    def mirrorDir(self, src_uri, dest_dir):
        ''' Mirror a remote HTTPS directory.
            calls 'wget --no-check-certificate --no-parent -nH --mirror'.
        '''
        p = path(urlparse.urlsplit(src_uri)[2]).splitall()
        # Deals with non-ending slash.
        # Non-ending clash url ends up with an empty string in the
        # last index of the list when a splitall is done
        if not p[-1]:
            cutaway = len(p[1:]) - 1
        else: # non-ending slash
            cutaway = len(p[1:])
            src_uri = src_uri + '/' # Append a trailing slash
        if cutaway < 0:
            cutaway = 0

        wget = subprocess.Popen(['wget', '--no-verbose', 
                                '--no-check-certificate', '--no-parent',
                                '--cut-dirs=%s' % cutaway, '-nH', 
                                '--mirror', src_uri,
                                '--directory-prefix', dest_dir],
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        retcode = wget.wait()
        decipherWgetErrors(src_uri, wget.stderr.readlines())
        return dest_dir 

class YOUFetchHandler(HTTPSFetchHandler):
    ''' HTTPSFetchHandler class subclasses from HTTPFetchHandler, and has
        identical function to its superclass, except that it handles the
        https protocol
    '''
    def __init__(self, **kwargs):
        HTTPSFetchHandler.__init__(self, **kwargs)

    def fetch(self, **kwargs):
        ''' Check if we are fetching a single file or
            mirroring a directory and act accordingly.
        '''
        if not checkUserPassFormat(kwargs['uri']):
            raise InvalidUserPassPairException, 'Invalid username/password given.'

        # you:// -> https://
        args = urlparse.urlsplit(kwargs['uri'])
        new_args = ['https']
        new_args.extend(args[1:])
        kwargs['uri'] = urlparse.urlunsplit(new_args)

        self.validateArgs(**kwargs)

        dest = self.downloadFile(kwargs['uri'],
                                 kwargs['destdir'])
        return (True, dest)


    def downloadFile(self, src_uri, dest_dir):
        ''' Download only one file from the src to destdir. '''
        # Obtain the destination filename after stripping away trailing Slash
        if not src_uri.endswith('/'):
            dest_filename = path(urlparse.urlsplit(src_uri)[2]).basename()
        else:
            dest_filename = path(urlparse.urlsplit(src_uri[-1])[2]).basename()
        if not dest_filename:
            raise FetchException,'Unable to obtain a destination file to download'
       
        dest_path = path(dest_dir) / dest_filename
        if dest_path.exists():
            dest_path.remove()
        p = subprocess.Popen(['wget', '--no-verbose',
                              '--no-check-certificate', src_uri,
                              '--directory-prefix', dest_dir],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        retcode = p.wait()
        stderr = p.stderr.readlines()
        if retcode:
            decipherWgetErrors(src_uri, stderr)
        return path(path(dest_dir) / dest_filename)
